# -*- coding: utf-8 -*-
from UserDict import UserDict
from abc import ABCMeta, abstractmethod
from bson import ObjectId
from flask import request
from flask_pymongo import DESCENDING, ASCENDING
from flask_dmango.mongodb_helpers import DmangoQueryBuilder, dmango_query


class AbstractRequestDmango(UserDict, dict):

    def __init__(self, req):
        assert req is request
        UserDict.__init__(self)
        self.args = req.args
        self.is_parse = False

    @abstractmethod
    def parse(self, args):
        raise Exception('parse logic override ')

    def lazy_parse(self):
        """
        요청이 없을때는 파싱하는 문제를 해결하기위해 파싱은 최대한 미룬다.
        @param args:
        @param kwargs:
        @return:
        """
        if not self.is_parse:
            self.parse(self.args)


    @property
    def sort(self):
        self.lazy_parse()
        return self.get('sort', None)

    @sort.setter
    def sort(self, value):
        self.__setitem__('sort', value)

    @property
    def page_num(self):
        self.lazy_parse()
        return self.get('page_num', None)

    @page_num.setter
    def page_num(self, value):
        self.__setitem__('page_num', value)

    @property
    def page_size(self):
        self.lazy_parse()
        return self.get('page_size', None)

    @page_size.setter
    def page_size(self, values):
        self.__setitem__('page_size', values)

    @property
    def start_idx(self):
        self.lazy_parse()
        return self.get('start_idx', None)

    @start_idx.setter
    def start_idx(self, values):
        self.__setitem__('start_idx', values)

    @property
    def end_idx(self):
        self.lazy_parse()
        return self.get('end_idx', None)

    @end_idx.setter
    def end_idx(self, values):
        self.__setitem__('end_idx', values)

    @property
    def match(self):
        self.lazy_parse()
        return self.get('match', None)

    @match.setter
    def match(self, values):
        self.__setitem__('match', values)

    @property
    def unwind(self):
        self.lazy_parse()
        return self.get('unwind', None)

    @unwind.setter
    def unwind(self, values):
        self.__setitem__('unwind', values)

    @property
    def group(self):
        self.lazy_parse()
        return self.get('group', None)

    @group.setter
    def group(self, values):
        self.__setitem__('group', values)

    @property
    def fields(self):
        self.lazy_parse()
        return self.get('fields', None)

    @fields.setter
    def fields(self, values):
        self.__setitem__('fields', values)

    @property
    def id(self):
        self.lazy_parse()
        return self.get('id', None)

    @id.setter
    def id(self, value):
        self.__setitem__('id', value)

    @property
    def options(self):
        self.lazy_parse()
        return self.get('options', None)

    @options.setter
    def options(self, value):
        self.__setitem__('options', value)


class RequestDmango(AbstractRequestDmango):
    def parse(self, args):
        self.update(parse_match(args))
        self.update(parse_sort(args))
        self.update(parse_offset(args))
        self.update(parse_group(args))
        self.update(parse_id(args))
        self.update(parse_options(args))
        self.update(parse_fields(args))


def split_generator(values):
    for line in values if isinstance(values, (tuple,list)) else tuple(values):
        for val in line.split(','):
            yield val


def parse_offset(args):
    _page_num = int(args.get('_page_num', 1))
    _page_size = int(args.get('_page_size', 20))
    _start_idx = int(args.get('_start_idx', (_page_num-1)*_page_size))
    _end_idx = int(args.get('_end_idx', (_page_num*_page_size)-1))
    return {'page_num': _page_num, 'page_size':  _page_size, 'start_idx': _start_idx, 'end_idx': _end_idx}


def parse_sort(args):
    _sort = []
    for sort_value in args.getlist('_sort'):
        t = sort_value.split('__')
        if len(t) == 2 and t[1] in ('desc','DESC'):
            _sort.append((t[0], DESCENDING))
        elif len(t[0]) > 0:
            _sort.append((t[0], ASCENDING))
    return {'sort': _sort}

find_field_end_idx_fn = lambda text, delis: min(text.find(d) if text.find(d) > 0 else len(text) for d in delis)

def parse_match(args):
    _query = DmangoQueryBuilder()
    exclude_fields = {k: True for k in split_generator(args.getlist('_exclude_query_fields'))}
    for dmango_fn in filter(lambda x: not x.startswith("_"), args):
        field_end_idx = find_field_end_idx_fn(dmango_fn,  ('__', ':', '~'))
        field_name = dmango_fn[:field_end_idx]
        if exclude_fields.has_key(field_name):
            continue # _exclude_query_fields 라면 제외
        _query.query(dmango_fn, args.getlist(dmango_fn))

    return {'match': _query()}


def parse_group(args):
    return {'unwind': split_generator(args.getlist('_unwind')), 'group': args.getlist('_group')}


def parse_id(args):
    return {'id': ObjectId(args.get('_id')) if args.get('_id') else None}


def parse_fields(args):
    _fields = {k: True for k in filter(lambda x: x and len(x) > 0, split_generator(args.getlist('_fields')))}
    return {'fields': _fields}


def parse_options(args):
    _options = {}
    for val in split_generator(args.getlist('_options')):
        t = val.split(":")
        if len(t)==2:
            _options[str(t[0])] = str(t[1].strip())
    return {'options': _options}



if __name__ == '__main__':
    pass


#-*- coding: utf-8 -*-
from functools import wraps
from flask import request, render_template_string, current_app

__templated_function_map = {}


def register(name):
    """flask에서 템플릿을 등록할때 사용한다.
    등록된 템플릿을 사용하는것은 @templated 이나 @after 를 사용한다.

    사용예제::

        @register('json')
        def json_template(result, count):
            rv = {'result': result, 'count': count}
            return json.dumps(my_json, ensure_ascii=False)

    @param name: 템플릿이름
    @return:
    """
    def decorator(f):
        __templated_function_map[name] = f
        print ' * View Decorators Register : name=%s, function=%s' % (name , str(f))
        return f
    return decorator


def templated(name=None, begin=None, end=None, exception=None):
    """register에서 등록한 것을 사용할때 이용한다.
    주의할 사항은, dict형태의 파라미터를 받아 string 결과를 만드는 함수가 등록되어 있어야 한다.

        사용예제:
            @app.route('/find')
            @templated('json')
            def find():
                rv = ['a','b','c']
                return dict(result=rv, count=len(rv))
    @param name: 템플릿 함수명을 의미한다.
    @param begin: 템플릿 함수를 실행하기전에 처리할 것을 의미한다 (파라미터로를 입력으로 받는다)
    @param end: 템플릿 함수를 실행한후 실핼할것을 의미한다 (결과를 입력으로 받는다)
    @param exception : 예외처리가 발생했을때 실행할것을 의미한다 (exception, +  파라미터정보를 입력으로 받는다)
    @return: str 타입의 결과
    """
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            before_fn = __templated_function_map.get(begin, None)
            after_fn = __templated_function_map.get(end, None)
            catch_fn = __templated_function_map.get(exception, None)
            render_template = __templated_function_map.get(name)
            if not render_template: raise Exception('not found template. (name=%s)' % (name, ))

            try:
                if before_fn: before_fn(*args, **kwargs)
            except Exception as e:
                print "[ERROR] @before execute failed.", e

            ## flask.route 실행
            try:
                ctx = f(*args, **kwargs)
            except Exception as e:
                catch_fn(exception=e, *args, **kwargs)

            try:
                if after_fn: after_fn(**ctx)
            except Exception as e:
                print "[ERROR] @after execute failed.", e

            if ctx is None:
                ctx = {}
            elif not isinstance(ctx, dict):
                return ctx
            text = render_template(**ctx)
            return render_template_string(text, **ctx)
        return decorated_function
    return decorator


def begin(name=None):
    """register에서 등록한 것을 사용할때 이용한다.
    이때, flask.route의 파라미터값을 그대로 사용한다.
    flask의 파라미터나 선처리를 할때 사용한다.

        사용예제:
            @app.route('/find')
            @begin('query_make')
            def find(param): ==> 이 결과 를 실행하기전에 선작업을 하고 싶을때 한다.
                rv = db.select(request.query)
                return dict(result=rv, count=len(rv))

    @param name:  등록한 함수 데잍레이터명을 의미한다.
    @return:
    """
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            fn = __templated_function_map.get(name)
            if not fn:
                raise Exception('not found template. (name=%s)' % (name ,))
            try:
                fn(*args, **kwargs)
            except Exception as e:
                print "[ERROR] @before execute failed.", e
            rv = f(*args, **kwargs)
            return rv
        return decorated_function
    return decorator


def end(name=None):
    """register에서 등록한 것을 사용할때 이용한다.
    이때, flask.route에서 return한 값을 파라미터로 받는다.
    flask의 결과를 후킹하는 용도로 사용된다.

        사용예제:
            @app.route('/find')
            @end('json')
            def find():
                rv = ['a','b','c']
                return dict(result=rv, count=len(rv)) ==> 이 결과 뒤에 뭔가 추가작업을 하고 싶을때 한다.

    result, args, kwargs
    @param name:  등록한 함수 데잍레이터명을 의미한다.
    @return:
    """
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            fn = __templated_function_map.get(name)
            if not fn:
                raise Exception('not found template. (name=%s)' % (name ,))
            rv = f(*args, **kwargs)
            try:
                fn(**rv)
            except Exception as e:
                print "[ERROR] @after execute failed.", e
            return rv
        return decorated_function
    return decorator


def exception(name=None):
    """register에서 등록한 것을 사용할때 이용한다.
    이때, 'exception=e' 형태의 값과 flask.route에서의 파라미터를 받는다.
    flask의 결과를 후킹하는 용도로 사용된다.

        사용예제:
            @app.route('/find')
            @exception('err_json')
            def find():
                rv = ['a','b','c']
                return dict(result=rv, count=len(rv)) ==> 이 결과 뒤에 뭔가 추가작업을 하고 싶을때 한다.

    @param name:
    @return:
    """
    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            fn = __templated_function_map.get(name)
            if not fn:
                raise Exception('not found template. (name=%s)' % (name ,))
            try:
                rv = f(*args, **kwargs)
            except Exception as e:
                fn(exception=e,*args, **kwargs)
            return rv
        return decorated_function
    return decorator


#-*- coding: utf-8 -*-
"""
    method 에 대한 처리를 한다.
    mongodb 의 오퍼레이션에 대한 처리를 위한 펑션이 모인곳으로 알면된다.
"""
import re as rex

from flask_dmango.query.exceptions import DmangoValueException


def __default_rex_search(f, v, func):
    if isinstance(v, list):
        return {f: {'$in': [func(_v) for _v in v]}}
    if isinstance(v, unicode) or isinstance(v, str):
        return {f: func(v)}
    raise TypeError("NOT Support Type.")


def like(f, v):
    return __default_rex_search(f, v, lambda _v: rex.compile(rex.escape(unicode(_v))))


def ignore_case_like(f, v):
    return __default_rex_search(f, v, lambda _v: rex.compile(rex.escape(unicode(_v)), flags=rex.IGNORECASE))


def startswith(f, v):
    return __default_rex_search(f, v, lambda _v: rex.compile('^' + rex.escape(unicode(_v))))


def ignore_case_startswith(f, v):
    return __default_rex_search(f, v, lambda _v: rex.compile('^' + rex.escape(unicode(_v)), flags=rex.IGNORECASE))


def endswith(f, v):
    return __default_rex_search(f, v, lambda _v: rex.compile(rex.escape(unicode(_v)) + '$'))


def ignore_case_endswith(f, v):
    return __default_rex_search(f, v, lambda _v: rex.compile(rex.escape(unicode(_v)) + '$', flags=rex.IGNORECASE))


def exact(f, v):
    return {f: v}


def gt(f, v):
    return {f: {'$gt': v}}


def gte(f, v):
    return {f: {'$gte': v}}


def lt(f, v):
    return {f: {'$lt': v}}


def lte(f, v):
    return {f: {'$lte': v}}


def not_equals(f, v):
    return {f: {'$ne': v}}


def contains(f, v):
    if not isinstance(v, list):
        return {f: {'$in': [v]}}
    return {f: {'$in': v}}


def not_contains(f, v):
    if not isinstance(v, list):
        return {f: {'$nin': [v]}}
    return {f: {'$nin': v}}


def between(f, v):
    if not isinstance(v, list) and not isinstance(v, tuple):
        raise TypeError('$betwwen method is support only list type. (error = "%s")' % (str(type(v)),))
    if len(v) != 2:
        raise DmangoValueException('$betwwen method is two value expected, but size = %d' % (len(v),))
    start_val = v[0] if v[0] < v[1] else v[1]
    end_val = v[0] if v[0] > v[1] else v[1]
    return {f: {'$gte': start_val, '$lte': end_val}}


METHOD_FUNCTION_DICT = {'like': like, 'ilike': ignore_case_like, 'gt': gt, 'lt': lt, 'gte': gte, 'lte': lte,
                        'in': contains, 'nin': not_contains, 'ne': not_equals, '': exact, 'startswith': startswith,
                        'endswith': endswith, 'istartswith': ignore_case_startswith, 'iendswith': ignore_case_endswith,
                        'between': between
                        }

if __name__ == '__main__':
    pass


# -*- coding: utf-8 -*-
import sys
reload(sys)
sys.setdefaultencoding('utf-8')
from flask import Flask, Blueprint
from flask_dmango import Dmango

app = Flask(__name__)
dmango = Dmango(app)
dmango.register_mongodb('server1', URI='mongodb://sample:sample@ds035240.mlab.com:35240/dmango')


bp = Blueprint('name1', __name__)
bp2 = Blueprint('name2', __name__)


@bp.route('/')
def test1():
    return "sample"


@bp2.route('/t2')
def test2():
    c = Dmango.find_collection('server1', db_name='dmango', collection_name='johnkim')
    for row in c.find(end_idx=10):
        print "-----------------------"
        for k in row:
            print k, row[k]
    return 'ok'

app.register_blueprint(bp)
dmango.register_blueprint(bp2)


if __name__ == '__main__':
    app.run(debug=True)


# -*- coding: utf-8 -*-
"""
    flask-dmango
    ~~~~~~~~~~~~

    이 모듈은  flask.ext.pymongo , request parser, restful-api  를 기본으로 제공한다.
    Dmango 를 통해 mongodb 를 등록하게 되면, 기본적인 restful-API 가 제공된다고 보면된다.
    특별한 코딩없이 바로 데이터를 저장/삭제/검색이 가능하다.
    그리고 기본으로 admin 기능도 제공되어 데이터 관리기능까지 제공된다.

    :copyright: (c) 2014 by mincheol jeong
    :license: BSD, see LICENSE for more details.
"""
__version__ = '0.0.3-dev'
from .mongodb_helpers import DmangoEngine, dmango_query, QueryBuilder, DmangoQueryBuilder
from .request_helpers import RequestDmango, AbstractRequestDmango
from .dmango import Dmango


#-*- coding: utf-8 -*-
from flask import request, Blueprint, render_template
dmango_help = Blueprint('dmango_help', __name__, template_folder='../templates', static_folder='../static_folder')

@dmango_help.route('/')
def index():
    return render_template('dmango/index.html')


#-*- coding: utf-8 -*-


class DmangoException(Exception):
    def __init__(self, message):
        self._message = message

    def _get_message(self):
        return self._message

    def _set_message(self, message):
        self._message = message

    message = property(_get_message, _set_message)


class NotSupportMethod(DmangoException):
    def __init__(self, method):
        self.method = method

    def __str__(self):
        return '"%s" is not a valid method name' % (str(self.method), )


class DmangoValueException(DmangoException):
    """
     value 를 관련 문제를 의미한다.
     예를 들어 'a' 를  int 로 변경하려고 했을경우라거나... 값이 허용되는게 아니라거나.
    """
    def __init__(self, value_name):
        self._value_name = value_name

    def __str__(self):
        return '[DMANGO-100002] value Error. (value name="%s")' % (self._value_name,)



class DmangoParseException(DmangoException):
    def __init__(self, errmsg='parse error'):
        self.errmsg = errmsg

    def __str__(self):
        return "[DMANGO-100003] " + self.errmsg



# -*- coding: utf-8 -*-
"""
    flask-dmango
    ~~~~~~~~~~~~

    이 모듈은  flask.ext.pymongo , request parser, restful-api  를 기본으로 제공한다.
    Dmango 를 통해 mongodb 를 등록하게 되면, 기본적인 restful-API 가 제공된다고 보면된다.
    특별한 코딩없이 바로 데이터를 저장/삭제/검색이 가능하다.
    그리고 기본으로 admin 기능도 제공되어 데이터 관리기능까지 제공된다.

    :copyright: (c) 2014 by mincheol jeong
    :license: BSD, see LICENSE for more details.
"""
from flask_dmango import RequestDmango
from flask_dmango.mongodb_helpers import DmangoEngine
from flask_pymongo import PyMongo
from flask import current_app, request, Blueprint


class Dmango(object):
    def __init__(self, app, root_url='/dmango'):
        """
        dmango 의 기본  base 를 지정한다.

        @param app: Flask's app
        @param root_url: default root_url is '/dmango'
        @return: none
        """
        self.__root_url = root_url
        self.__app = app
        self.init_app(app)


    @staticmethod
    def find_mongodb(server_name):
        """
        서버명으로 Pymongo 객체를 찾는다.
        특이한건  current_app 의 드망고의 서버이름에서 찾기 때문에,  app 별로 관리된다.
        @param server_name: 등록한 서버명
        @return: 파이몽고 객체
        @rtype: PyMongo
        """
        if 'dmango' not in current_app.extensions:
            raise Exception('not exist app.extentions.dmango')
        return current_app.extensions['dmango'][server_name]

    @staticmethod
    def find_collection(server_name, db_name, collection_name):
        """ 드망고에 등록된 몽고디비정보를 찾아, 드망고 콜렉션 오브젝트를 받는다.
        find_mongodb에서 좀더 쉽게 콜렉션을 리턴받기위해 사용하는 메소드이다.

        @param server_name: 등록한 몽고디비 서버정보(서버명)
        @param db_name: 디비명
        @param collection_name: 콜렉션명
        @return: Dmango의 Collection객체
        @rtype: DmangoEngine
        """
        return DmangoEngine(Dmango.find_mongodb(server_name), db_name, collection_name)

    def init_app(self, app):
        if 'dmango' not in app.extensions:
            app.extensions['dmango'] = dict()  # {mongo_name: pymongo}

        print '[DMANGO] restfulAPI loading...'
        from flask_dmango.view.dmango_restful import dmango_restful
        self.register_blueprint(dmango_restful, url_prefix=self.__root_url)
        print '[DMANGO] help loading...'
        from flask_dmango.view.dmango_help import dmango_help
        self.register_blueprint(dmango_help, url_prefix=self.__root_url+"_help")
        print '[DMANGO] Admin Blueprint loading...'
        from flask_dmango.view.dmango_admin import dmango_admin
        self.register_blueprint(dmango_admin, url_prefix=self.__root_url+"_admin")


    def register_mongodb(self, server_name, **options):
        """
        몽고디비의 정보를 등록한다.
        flask.ext.pymongo 를 이용하여 등록하므로,  app.config 에  options 가 등록된다.
        이때 server_name 이 헤더가 되어 등록되어 진다.
        그러므로,  app 별로  server_name 이름이 겹쳐지는것은  허용되지 않는다.

        @param server_name: 나중에 몽고디비 접근정보를 찾기위한  key 로 사용된다.
        @param options:  mongodb 의 접속 정보를 의미한다.  URI=..., PORT..
        @return: None
        """
        if not server_name:
            raise Exception('name is not empty.')
        if options:
            key = lambda o: '%s_%s' % (server_name, o)
            print "[DAMNGO] mongodb connection info, " , {key(k): options[k] for k in options}
            self.__app.config.update({key(k): options[k] for k in options})
        try:
            if server_name in self.__app.extensions['dmango']:
                raise Exception('duplicate dmango name. "%s"' % server_name)
            self.__app.extensions['dmango'][server_name] = PyMongo(self.__app, server_name)
        except Exception as e:
            raise Exception('connection failed. (%s)' % (e,))

    def register_blueprint(self, blueprint, **option):
        assert blueprint , Blueprint
        """
        app.register 의 기능을 대행해 준다.
        다른점은  dmango 의  reqeust를 파싱한 결과가 기본으로 붙여준다는점만 다르다.
        flask 의  reqeust 객체에 다음과 같은 정보가 붙는점이 다르다. (불필요하면  app 에서 바로  blueprint 를 붙이도록 한다)

        예제)
            request.dmango_query   :   builder.Query 객체를 리턴함.
            request.dmango_options :   builder.Option  객체를 리턴함.

        @param blueprint: 등록할 블루프린터를 의미한다.
        @param option: app.register_blueprint 와 동일한 옵션을 의미한다.
        @return: none
        """
        def __bind_request_query():
            setattr(request, 'dmango', RequestDmango(request))
        blueprint.before_request(__bind_request_query)
        self.__app.register_blueprint(blueprint, **option)



# -*- coding: utf-8 -*-
from collections import namedtuple

from flask_dmango.query.exceptions import NotSupportMethod
from flask_dmango.query.methods import METHOD_FUNCTION_DICT
from flask_dmango.query.variables import DATA_TYPE_SYNTAX


def to_cast(values, type_name=None, convert_list=False):
    """
        드망고에 정의된 타입정보를 이용해  values 를 변환해 준다.
        tip: <필드명>__<메소드명>[]:<type_name>=value & ....
    @param values:  값을 의미한다. 내부적으로 [] 인 값과 단일값을 알아서 체크해 준다.
    @param type_name: 타임명을 의미한다. 드망고의 :<type_name> 을 의미한다.
    @param convert_list: 드망고에서 리스트 타입을 사용자가 강제화 한 경우를 의미한다. []를 강제화 하여 사용한 경우이다.
    예를 들면,  img[]=1_jpg&img[]=2_jpg ... 이때는 1개일때도 강제로 list 의 값으로 자동변환된다.
    @return:
    """
    val = None
    if isinstance(values, list):
        length = len(values)
        if length <= 0:
            return None
        if length > 1:
            val = values
        if length == 1 and convert_list:
            val = values
        if length == 1 and not convert_list:
            val = values[0]
    else:
        val = [values] if convert_list else values

    try:
        if not type_name:
            return val
        return map(lambda v:DATA_TYPE_SYNTAX[type_name](v), val) if isinstance(val, list) else DATA_TYPE_SYNTAX[type_name](val)
    except Exception as e:
        raise TypeError('unknown type name "%s" <%s>' % (type_name,str(e)))


def to_method_value(_field_name, _values, _method_name=None, _sub_map_key=None):
    """
    메소드에 따라 쿼리를 생성한다.
    @param _field_name: 필드명
    @param _values: 값
    @param _method_name: 메소드명을 의미한다.  in , nin, istartswith  같은 값을 의미한다.
    @param _sub_map_key :  '필드명[서브맵키]' 형태로 정의된것으로 dict 로 만들어 져야 할 경우를 의미한다.
    @return:
    """
    if _values is None or _values == '': return None
    mn = _method_name.lower() if _method_name else ''
    if mn in METHOD_FUNCTION_DICT:
        if _sub_map_key:
            return {_field_name: METHOD_FUNCTION_DICT[mn](_sub_map_key, _values)}
        else:
            return METHOD_FUNCTION_DICT[mn](_field_name, _values)
    raise NotSupportMethod(_method_name)


def __min_idx(*args):
    f = filter(lambda x: x >= 0, args)
    return min(f) if f else 0


def __find_sidx(s, find_str, default_val=-1):
    t = s.find(find_str)
    if t < 0: return default_val
    t += len(find_str)
    return t if t > 0 else default_val


def field_name(token=''):
    end_idx = __min_idx(token.find('__'), token.find(':'), token.find('~'), len(token))
    if token == 0: raise TypeError('empty value.')
    if end_idx <= 0: raise TypeError('"field_name" not found. (=' + str(token) + ')')

    # [OK]
    return token[:end_idx]


def field_sub_info(field_name=''): # return (필드명,<서브필드명[|None]>,<타입[list|dict|None]>)
    if field_name.endswith('[]'):
        return field_name[:-2], None, 'list'
    if field_name.find('[') > 0 and field_name.endswith(']'):
        main_key = field_name[:field_name.find('[')]
        sub_key = field_name[field_name.find('[') + 1:-1]
        return main_key, sub_key, 'dict'
    return field_name, None, None



def method_name(token=''):
    start_idx = __find_sidx(token, '__')
    if start_idx < 0: return ''
    end_idx = __min_idx(token.find(':'), token.find('~'), len(token))
    method = token[start_idx:end_idx]
    return '' if method == 'exact' else method


def type_name(token=''):
    start_idx = __find_sidx(token, ':')
    if start_idx<0: return ''
    if token.find('__',start_idx)>=0:
        raise TypeError('[DMANGO-100009] "type" is next "method" unexpected. (=' + str(token) + ')')
    end_idx = __min_idx(token.find('~',start_idx), len(token))
    f = token[start_idx:end_idx]
    return f if len(f) > 0 else None



def nick_name(token=''):
    start_idx = __find_sidx(token, '~')
    if start_idx < 0 : return ''
    if max(token.find('__',start_idx), token.find(':',start_idx)) > 0:
        raise TypeError('[DMANGO-100009] "nickname" is next "method" or "type" unexpected. (=' + str(token) + ')')
    return token[start_idx:]


DmangoFieldInfo = namedtuple('DmangoFieldInfo', ['field_name','sub_field_name','method_name','type_name','sub_type_name','nick_name'])


def to_dmango_field_info(dmango_field_name):
    """
    드망고의 필드명을 정보로 쪼갠다.

    @param dmango_field_name:
    @return:
    """
    _field_name, _sub_field_name, _sub_type_name = field_sub_info(field_name(dmango_field_name))
    _method_name = method_name(dmango_field_name)
    _type_name = type_name(dmango_field_name)
    _nick_name = nick_name(dmango_field_name)
    return DmangoFieldInfo(field_name=_field_name, sub_field_name=_sub_field_name,
                           method_name=_method_name, type_name=_type_name,
                           sub_type_name=_sub_type_name, nick_name=_nick_name)


def to_variable(dmango_field_info, values):
    """
    드망고 필드정보를를 이용한 값을 하나 생성한다.
    @param dmango_field_info: DmangoFieldInfo  타입정보를 의미한다.
    @param values:
    @return: {필드명: 값}  or {필드명 : {메소드: 값...}}
    """
    try:
        df = dmango_field_info
        value = to_cast(values, df.type_name, 'list' == df.sub_type_name)
        return to_method_value(df.field_name, value, df.method_name, df.sub_field_name)
    except TypeError as e:
        raise Exception('[DMANGO-100022] value cast error. (type=%s, values=%s)' % (str(df.type_name), str(values)))
    except Exception as e:
        if not isinstance(dmango_field_info, DmangoFieldInfo):
            raise TypeError('[DMANGO-100021] argument type error. (extected type = "DmangoFieldInfo")')
        if not df.type_name in DATA_TYPE_SYNTAX:
            raise Exception('[DMANGO-100020] unknown "field_type" name. (you=%s, extected="%s")' %
                            (str(df.type_name) , str(DATA_TYPE_SYNTAX.keys()),))
        raise Exception('[DMANGO-100023] maybe crazy variable error. %s' % str(e))




if __name__ == '__main__':
    print ">>>>> to_cast : start "
    print '01', type(to_cast('1234')) == str
    print '02', type(to_cast('1234','int')) == int
    print '03', type(to_cast(['1234'],'int')) == int
    print '04', to_cast(['1234'],'list') == ['1234'] # :list == > [1234]
    print '05', to_cast('1234','list') == ['1234'] # :list == > [1234]
    print '06', to_cast(['1234']) == '1234' # == > 1234
    print '07', to_cast(['1234,5678'],'list') == ['1234','5678'] # :list, 1234,5678 -> ['1234','5678']
    print '08', to_cast(['1234','5678'], 'int', True) == [1234, 5678] # == > [1234,5678]
    print '09', to_cast(['12,34','56,78'], 'list', True) == [['12','34'],['56','78']] # == > [['12','34'],['56','78']]
    print '10', to_cast(['12,34'], 'list', True) == [['12','34']]# == > [['12','34']]
    print '11', to_cast('12,34', 'list') == ['12','34']# == > ['12','34']
    print '12', to_cast([], 'list') == None# == > None
    print '13', to_cast([], 'list', True) == None # == > None
    print '14', to_cast('A', None, True) == ['A'] # == > [A]
    print '15', to_cast(['A'], None, True) == ['A'] # == > [A]
    print '16', to_cast(['123','456'], 'str', True) == ['123','456'] # []:str ==> ['123','456']
    print '17', to_cast(['{"name":"jmc", "age":33}'], 'json', True) == [{u'name':u'jmc', u'age':33}]# []:json ==> [{name:jmc, age:33}]
    print '18', to_cast(['{"name":"jmc", "age":33}'], 'json', False) == {u'name':u'jmc', u'age':33}# []:json ==> {name:jmc, age:33}
    print '19', to_cast(['1234','555'],'int') == [1234, 555]# :int ...
    print ">>>>> to_cast : end "
    print ""

    print ">>>>> to_one_vaue : start "
    print "01", to_method_value("title",'GIL_DONG')
    print "02", to_method_value("title",'GIL_DONG', 'in')
    print "03", to_method_value("user",'GIL_DONG', 'in', 'title')
    print "04", to_method_value("user",['GIL_DONG','HIHI'], 'in', 'title')
    print "05", to_method_value("user",['GIL_DONG','HIHI'], 'like')
    print ">>>>> to_one_value : end "
    print ""

    print ">>>>> to_dmango_field_info : start "
    print "01", to_dmango_field_info("title[sub]:int~tag")
    print "02", to_dmango_field_info("title")
    print "03", to_dmango_field_info("title:list")
    print "04", to_variable(to_dmango_field_info("title:list"), 'jmc')
    print ">>>>> to_dmango_field_info : end "
    print "end"



# -*- coding: utf-8 -*-
import datetime as datetime_lib
from flask_pymongo import PyMongo
from pymongo import MongoClient, MongoReplicaSetClient
from flask_dmango.query.fields import to_variable, to_dmango_field_info


class QueryBuilder:
    def __init__(self):
        self._query = {}

    def exist_and(self):
        return '$and' in self._query

    def is_empty(self):
        return len(self._query) <= 0

    def append(self, query):
        if query is None:
            return self
        if isinstance(query, QueryBuilder):
            query = query()
        if not isinstance(query, dict):
            raise TypeError('append method type is [dict]')

        for k in query:
            key = k
            value = query.get(k)
            if self.is_empty():
                self._query.update({key: value})
            elif self.exist_and():
                is_update = False
                for row in self._query['$and']:
                    if QueryBuilder.multi_check_and_update(key, value, row):
                        is_update = True
                        break
                if not is_update:
                    self._query['$and'].append({key: value})
            elif not self.exist_and():
                if not QueryBuilder.multi_check_and_update(key, value, self._query):
                    self._query = {'$and': [self._query, {key: value}]}
        return self

    def clear(self):
        self._query.clear()

    """
    price : {$gt:100}
    price : {$gt:100, dd}
    """
    @staticmethod
    def multi_check_and_update(key, value, m):
        if isinstance(m, dict) and key not in m: # 해당안에 겹치는 키가 업으면 기냥 깽겨넣음.
            m.update({key: value})
            return True
        if isinstance(value, dict) and isinstance(m.get(key), dict) and len(value) == 1:
            # 해당안에 겹쳐지는걸 찾으면 그안에 낑겨서  update.
            my_method = value.keys()[0]
            if not my_method in m[key]:
                m[key].update(value)
                return True
        return False

    def __call__(self):
        return self._query if len(self._query) > 0 else None

    def __str__(self):
        return self._query.__str__()

    def __len__(self):
        return len(self._query)


class DmangoQueryBuilder:
    def __init__(self):
        self._query = QueryBuilder()

    def clear(self):
        self._query.clear()

    def append(self, query):
        if isinstance(query, (QueryBuilder,DmangoQueryBuilder)):
            self._query.append(query())
        elif isinstance(query, dict):
            self._query.append(query)
        else:
            raise TypeError('append method type is [dict | QueryBuilder | DmangoQueryBuilder]')

    def query(self, dmango_field_name, values, default_value=None):
        self._query.append(dmango_query(dmango_field_name, values, default_value))
        return self

    def __call__(self, *args, **kwargs):
        return self._query()

    def __len__(self):
        return len(self._query)

    def __str__(self):
        return str(self._query())


class DmangoEngine:
    """
    드망고에서 관리되는 몽고디비 실행 엔진이다.
    """
    def __init__(self, connection, db_name=None, collection_name=None):
        if not isinstance(connection, (PyMongo, MongoClient, MongoReplicaSetClient)) or not connection:
            raise TypeError('connection must be a class of PyMongo | MongoClient | MongoRepliaceSetClient.')
        self.__connection = connection
        self.__db_name = None
        self.__collection_name = None
        self.change(db_name, collection_name)

    def change(self, db_name=None, collection_name=None):
        if db_name:
            self.__db_name = db_name
        if collection_name:
            self.__collection_name = collection_name
        return self

    @property
    def connection(self):
        return self.__connection

    @connection.setter
    def connection(self, value):
        self.__collection_name = value

    @property
    def collection(self, db_name=None, collection_name=None):
        _db_name = db_name or self.__db_name
        _cn_name = collection_name or self.__collection_name
        if not _cn_name:
            raise TypeError('collection name is empty.')

        if isinstance(self.__connection, PyMongo):
            return self.__connection.cx[_db_name][_cn_name] if _db_name else self.__collection_name.db[_cn_name]
        else:
            return self.__connection[_db_name][_cn_name]

    def find(self, match={}, sort=None, start_idx=0, end_idx=20, fields=None):
        c = self.collection

        if fields:
            fields = {f:True for f in fields}

        c = c.find(match, projection=fields)

        if not c:
            return list()
        if sort:
            c = c.sort(sort)

        if end_idx >= 0:
            c = c[start_idx:end_idx+1]
        elif end_idx < 0:
            raise ValueError('ValueError: page_num or page_size by <= 0')
        else:
            c = c[start_idx:]
        return c

    def find_one(self, match, sort=None, fields=None):
        c = self.collection
        c = c.find_one(match, fields=fields if fields else None)
        if sort:
            c = c.sort(sort)
        return c

    def string_group(self, match, group_fields, limit=100, unwind_fields=None, filter_query=None):
        c = self.collection
        match_query = [{k: {'$exists': 'true'}
                                           for k in (group_fields
                                                     if isinstance(group_fields, (list, tuple))
                                                     else (group_fields,))}]
        if match:
            match_query.append(match)
            q = [{'$match': {'$and': match_query}}]
        else:
            q = [{'$match': match_query[0]}]

        for uw in unwind_fields if unwind_fields else tuple():
            q.append({'$unwind': '$%s' % uw})
        q.append({'$group': {'_id': {k: '$%s' % (k,)
                          for k in (group_fields
                                    if isinstance(group_fields, (list, tuple))
                                    else (group_fields,))}, 'count': {'$sum': 1}}})
        if filter_query:
            q.append(filter_query)
        q.append({'$sort': {'count': -1}})
        q.append({"$limit": limit})
        rv = c.aggregate(q)
        if u'result' in rv:
            rv = rv.get(u'result', tuple())

        for row in rv:
            row.update(row[u'_id'])
            del row[u'_id']
        return rv

    def count(self, match):
        c = self.collection
        c = c.find(match).count()
        return c

    def insert(self, data_json, w=1):
        """
        입력모드
        :param data_json: 저장할 json
        :param w: 0 - 빠르지만 안정성보장 안함, 1-어느정도 안정성 보장
        :return:
        """
        c = self.collection
        return c.insert(data_json if not isinstance(data_json, (list,)) else [data_json], w=w)

    def save(self, data_json):
        c = self.collection
        return c.save(data_json)

    def update(self, _id, set_json):
        c = self.collection
        return c.update({'_id': _id}, {'$set': set_json})

    def remove(self, _id):
        c = self.collection
        return c.remove({'_id': _id})

    def inc(self, _id, data_json):
        c = self.collection
        return c.update({'_id': _id}, {'$inc': {k: int(data_json.get(k, 1)) for k in data_json}})

    def push(self, _id, data_json):
        c = self.collection
        return c.update({'_id': _id}, {'$push': data_json})


def convert_dmango_special_values(values):
    today = datetime_lib.date.today()
    con_fn_dict = {
        "$sysdatetime": lambda: datetime_lib.datetime.today(),
        "$sysdate": lambda: today,
        "$today": lambda: "%d%02d%02d" % (today.year, today.month, today.day),
    }
    if not values:
        return None
    if isinstance(values, (str, unicode)) and str(values) in con_fn_dict:
        return con_fn_dict[str(values)]()
    if isinstance(values, (list, tuple)):
        for idx, row in enumerate(values):
            if isinstance(row, (str,unicode)) and str(row) in con_fn_dict:
                values[idx] = con_fn_dict[str(row)]()
    return values


find_field_end_idx_fn = lambda text, delis: min(text.find(d) if text.find(d) > 0 else len(text) for d in delis)

def dmango_query(dmango_field_name, values, default_value=None):
    if not isinstance(values, (int, float, long)):
        values = values if values and len(values) > 0 else default_value

    field_end_idx = find_field_end_idx_fn(dmango_field_name,  ('__', ':', '~'))
    field_name = dmango_field_name[:field_end_idx]
    if field_name.find(",") < 0:
        query = to_variable(to_dmango_field_info(dmango_field_name), convert_dmango_special_values(values))
    else:  # 콤마로 구분된 쿼리 구조 or 쿼리를 처리한다. (예: title,name=키워드 )
        tail_info_str = dmango_field_name[field_end_idx:]
        _sub_q = []
        for f in field_name.split(","):
            _q = to_variable(to_dmango_field_info(f+tail_info_str), convert_dmango_special_values(values))
            if _q:
                _sub_q.append(_q)
        query = {"$or": _sub_q} if len(_sub_q) > 0 else None

    return query


if __name__ == '__main__':
    b = DmangoQueryBuilder()
    b.query("title", ['apple','ttt'])
    b.query("price__gte:int", 5000)
    b.query("name,jmc__ilike", "jmc")
    print b




from setuptools import find_packages, setup
setup(
    name='Flask-Dmango',
    version='0.0.5',
    url='https://github.com/jungkoo/flask-dmango',
    license='BSD',
    author='jeong mincheol',
    author_email='deajang@gmail.com',
    description='Contents managements support for Flask + MongoDB applications',
    long_description=__doc__,
    zip_safe=False,
    platforms='any',
    packages=find_packages(),
    package_data={'':['templates/admin/*.html', 'templates/dmango/*.html']},
    install_requires=[
        'Flask >= 0.',
        'pymongo >= 3.2.2',
        'Flask-PyMongo >= 0.3.0',
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    # setup_requires=['nose'],
    # tests_require=['nose', 'coverage'],
    # test_suite='nose.collector',
)



#-*- coding: utf-8 -*-
from functools import wraps
from flask import current_app, request
from bson.json_util import dumps
import time

current_milli_time = lambda: int(round(time.time() * 1000))

def dmango_api_template(api_name, indent=None):
    """
    dict형태를 json형태로 리턴하는 템플릿
    @param api_name: api명
    @return:
    """

    def decorator(f):
        @wraps(f)
        def decorated_function(*args, **kwargs):
            start_time = current_milli_time()
            seed_result = {'api_name': api_name, 'msg': '성공', 'status': 200}
            try:
                ctx = f(*args, **kwargs) or dict()  ## flask.route 실행
                seed_result.update(**ctx)
            except Exception as e:
                seed_result['msg'] = '실패 (%s)' % (e,)
                seed_result['status'] = 404
                ctx = {}
            seed_result['time(ms)'] = current_milli_time() - start_time

            _indent = indent
            if hasattr(request, 'dmango') and not _indent:
                _indent = 2 if request.dmango.options.get('indent') in ('true', 'True', 'TRUE', 'YES', '1', 'Y') else None

            return current_app.response_class(dumps(seed_result, indent=_indent, ensure_ascii=False),
                                              mimetype='text/json;charset=UTF-8')
        return decorated_function
    return decorator


#-*- coding: utf-8 -*-
from bson.objectid import ObjectId
from flask import request, Blueprint, render_template
from flask_dmango import Dmango, RequestDmango

dmango_admin = Blueprint('dmango_admin', __name__, template_folder='../templates')

@dmango_admin.route("/")
def index():
    return render_template('admin/authors.html', test='testdddd')

@dmango_admin.route("/summary")
def summary():
    test = '1'
    return render_template('', test=test)

@dmango_admin.route("/server")
def server_list():
    return 'server_list'

@dmango_admin.route("/<server>.<db_name>.<collection_name>/list")
def contents_list(server, db_name, collection_name):
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    if q.id:
        rv = mongo.find(match={'_id': q.id})
        head_list = mongo.find(match={'_id': q.id})
        count = mongo.count(match={'_id': q.id})
    else:
        rv = mongo.find(match=q.match, sort=q.sort, start_idx=q.start_idx, end_idx=q.end_idx)
        head_list = mongo.find(match=q.match, sort=q.sort, start_idx=q.start_idx, end_idx=q.end_idx)
        count = mongo.count(match=q.match)
    # return dict(total=count,  page_num=q.page_num, page_size=q.page_size, result=rv)
    header = dict()
    for row in head_list:
        header.update({k: True for k in row})

    return render_template('admin/contents_list.html', header=header.keys(), result=rv, total=count,
                           page_size=q.page_size, page_num=q.page_num, max_page=int(count/q.page_size)+1)

@dmango_admin.route("/<server>.<db_name>.<collection_name>/detail/<_id>")
def detail(server, db_name, collection_name, _id):
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    doc = mongo.find_one({'_id': ObjectId(_id)})
    return render_template('admin/detail.html', data_types={}, result=doc if doc else dict())



#-*- coding: utf-8 -*-
import json as json_lib
from bson.objectid import ObjectId
import datetime as datetime_lib


def empty_value_none(func):
    def inner(*args, **kwargs):
        if len(args) >=0 and (args[0] is None or "" == args[0]):
            return None
        return func(*args, **kwargs)
    return inner


# 정의된  obj_type 이면 기냥  by pass  한다.
def same_object_by_pass(obj_type):
    def inner(f):
        def warpped_f(*args, **kwargs):
            if isinstance(args[0], obj_type):
                return args[0]
            else:
                return f(*args, **kwargs)
        return warpped_f
    return inner



@empty_value_none
@same_object_by_pass(datetime_lib.date)
def to_date_time(s):
    s = str(s)
    length = len(s)
    year = month = day = hour = minute = sec = 0
    if length < 8:
        raise TypeError('datetime length is short. (>=8, value: ' + s + ')')
    if length >= 4: year = int(s[:4])
    if length >= 6: month = int(s[4:6])
    if length >= 8: day = int(s[6:8])
    if length >= 10: hour = int(s[8:10])
    if length >= 12: minute = int(s[10:12])
    if length >= 14: sec = int(s[12:14])

    if month > 12: raise TypeError('month error. month: ' + str(month) + ' ,datetime:' + s)
    if day > 31: raise TypeError('day error. day: ' + str(day) + ' ,datetime:' + s)
    if hour > 24: raise TypeError('day error. hour: ' + str(hour) + ' ,datetime:' + s)
    if minute > 59: raise TypeError('day error. hour: ' + str(minute) + ' ,datetime:' + s)
    if sec > 59: raise TypeError('sec error. day: ' + str(sec) + ' ,datetime:' + s)

    return datetime_lib.datetime(year, month, day, hour, minute, sec)


def to_json(s, check_type=None):
    if s is None: return None
    try:
        tmp_json = json_lib.loads(s)
        if not check_type or isinstance(tmp_json, check_type):
            return tmp_json
    except ValueError as e:
        raise TypeError('[ERROR] json parse error !!! ==> %s' % (s,))
    raise TypeError('[ERROR] json parse result type error. (value=' + str(s) + ', result_type=' + str(type(tmp_json)) +
                    ', expected_type=' + str(check_type))

@empty_value_none
@same_object_by_pass(list)
def to_list(s):
    if not s.startswith("[") or not s.endswith("]"):  # json 형태의  list 는 아니지만  ,  를 구분자로 하는 값이 존재한다.
        return s.split(",") if ',' in s else [s]
    return to_json(s, list)

@empty_value_none
@same_object_by_pass(dict)
def to_dict(s):
    return to_json(s, dict)

@empty_value_none
@same_object_by_pass(int)
def to_int(s):
    return int(s)

@empty_value_none
@same_object_by_pass(long)
def to_long(s):
    return long(s)

@empty_value_none
@same_object_by_pass(float)
def to_float(s):
    return float(s)

@empty_value_none
@same_object_by_pass(str)
def to_string(s):
    return str(s)


DATA_TYPE_SYNTAX = {'int': to_int, 'long': to_long, 'float': to_float, 'string': to_string, 'text': to_string,
                    'str': to_string, 'json': to_json, 'datetime': to_date_time, 'list': to_list, 'dict': to_dict,
                    'oid': ObjectId, '': to_string}

if __name__ == '__main__':
    pass



#-*- coding: utf-8 -*-
from flask import request, Blueprint, url_for, redirect
from flask_dmango import Dmango
from flask_dmango.request_helpers import RequestDmango
from flask_dmango.view import dmango_api_template

dmango_restful = Blueprint('dmango_restful', __name__)


@dmango_restful.route('/<server>.<db_name>.<collection_name>/xxx', methods=['GET', 'POST'])
@dmango_api_template(api_name='xxx', indent=True)
def xxx(server=None, db_name=None, collection_name=None):
    return dict(match='xxx')


@dmango_restful.route('/<server>/', methods=['GET', 'POST'])
@dmango_api_template(api_name='show_dbs')
def show_dbs(server):
    db_list = Dmango.find_mongodb(server).cx.database_names()
    return dict(server=server, total=len(db_list), db_names=db_list)

@dmango_restful.route('/<server>.<db_name>/', methods=['GET', 'POST'])
@dmango_api_template(api_name='show_collecitons')
def show_collections(server, db_name):
    collection_list = Dmango.find_mongodb(server).cx[db_name].collection_names()
    return dict(server=server, total=len(collection_list), db_names=collection_list)


@dmango_restful.route('/<server>.<db_name>.<collection_name>/', methods=['GET', 'POST'])
@dmango_restful.route('/<server>.<db_name>.<collection_name>/api_list', methods=['GET', 'POST'])
@dmango_api_template(api_name='api_list')
def api_list(server, db_name, collection_name):
    url = lambda x: "./%s.%s.%s/%s" % (server, db_name, collection_name, x)
    contents = lambda name, comment: dict(name=name, url=url(name), comment=comment)
    r = (
        contents('param', '파라미터가 올바르게 만들어 지는지 테스트하기 위해서 사용'),
        contents('find_one', '한건만 검색할때 사용한다. 단, _id 조건이 있으면 나머지는 무시됨'),
        contents('find', '일반적인 검색을 할 때 사용한다.'),
        contents('count', '검색결과가 몇건인지 확인하기위해 사용'),
        contents('insert', '데이터를 저장하기위해 사용한다. 단, 드망고의 명령어는 사용하지 말자'),
        contents('save','데이터를 저장하기위해 사용한다. '
                                'insert 와의 차이점은 _id 가 허용되며, 같은 _id를 만나면  update 처리됨'),
        contents('update', '_id로 1건의 데이터만 찾아 update 가 가능하다.'),
        contents('remove', '_id 로 찾은 데이터를 삭제한다.'),
        contents('push', '_id 로 찾은 데이터에서  list 타입에 값을 하나 append 할때 사용'),
        contents('inc', '_id 로 찾은 데이터에서  넘겨받은 필드와 숫자값만큼 증가시켜준다.'),
        contents('string_group', '특정 필드를 그룹핑하여 카운팅해준다. _group= 이라는 값을 사용한다.'),
    )
    return dict(total=len(r), result=r)


@dmango_restful.route('/<server>.<db_name>.<collection_name>/find_one', methods=['GET', 'POST'])
@dmango_api_template(api_name='find_one')
def find_one(server, db_name, collection_name):
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    if q.id:
        rv = mongo.find_one(match={'_id': q.id}, fields=q.fields)
        count = mongo.count(match={'_id': q.id})
    else:
        rv = mongo.find_one(match=q.match, fields=q.fields)
        count = mongo.count(match=q.match)
    return dict(total=count, result=rv)


@dmango_restful.route('/<server>.<db_name>.<collection_name>/param', methods=['GET', 'POST'])
@dmango_api_template(api_name='param', indent=True)
def index(server=None, db_name=None, collection_name=None):
    q = request.dmango
    return dict(match=q.match, sort=q.sort, unwind=q.unwind, group=q.group,
                start_idx=q.start_idx, end_idx=q.end_idx, page_num=q.page_num, page_size=q.page_size,
                server=server, db=db_name, collection=collection_name, fields=q.fields, options=q.options
                )

@dmango_restful.route('/<server>.<db_name>.<collection_name>/find', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_find')
def find(server, db_name, collection_name):
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    if q.id:
        rv = mongo.find(match={'_id': q.id}, fields=q.fields)
        count = mongo.count(match={'_id': q.id})
    else:
        rv = mongo.find(match=q.match, sort=q.sort, fields=q.fields, start_idx=q.start_idx, end_idx=q.end_idx)
        count = mongo.count(match=q.match)
    return dict(total=count,  page_num=q.page_num, page_size=q.page_size, result=rv)

@dmango_restful.route('/<server>.<db_name>.<collection_name>/count', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_count')
def count(server, db_name, collection_name):
    q = request.dmango
    if q.id:
        raise Exception('find method is "_id" value not support.')
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    count = mongo.count(match=q.match)
    return dict(total=count)

@dmango_restful.route('/<server>.<db_name>.<collection_name>/insert', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_insert')
def insert(server, db_name, collection_name):
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    rv = mongo.insert(q.match)
    return dict(result=rv)

@dmango_restful.route('/<server>.<db_name>.<collection_name>/save', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_save')
def save(server, db_name, collection_name):
    q = request.dmango
    q = q.match.update(q.id)
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    rv = mongo.save(q)
    return dict(result=rv)

@dmango_restful.route('/<server>.<db_name>.<collection_name>/update', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_update')
def update(server, db_name, collection_name):
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    rv = mongo.update(q.id, q.match)
    return dict(result=rv)

@dmango_restful.route('/<server>.<db_name>.<collection_name>/push', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_push')
def push(server, db_name, collection_name):
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    rv = mongo.push(q.id, q.match)
    return dict(result=rv)

@dmango_restful.route('/<server>.<db_name>.<collection_name>/inc', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_inc')
def inc(server, db_name, collection_name):
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    rv = mongo.inc(q.id, q.match)
    return dict(result=rv)

@dmango_restful.route('/<server>.<db_name>.<collection_name>/remove', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_remove')
def remove(server, db_name, collection_name):
    q = request.dmango
    if not q.id:
        raise Exception('_id값이 비어있습니다')
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    rv = mongo.remove(q.id)
    return dict(result=rv)


@dmango_restful.route('/<server>.<db_name>.<collection_name>/string_group', methods=['GET', 'POST'])
@dmango_api_template(api_name='dmango_string_group')
def string_group(server, db_name, collection_name):
    """
    http://localhost:5000/dmango/server1.sada9_dev.testdb/string_group?_group=cate_nm&_unwind=cate_nm&_page_num=2&_page_size=5
    @param server:
    @param db_name:
    @param collection_name:
    @return:
    """
    q = request.dmango
    mongo = Dmango.find_collection(server, db_name=db_name, collection_name=collection_name)
    limit = q.end_idx + 1
    rv = mongo.string_group(q.match, group_fields=q.group, limit=limit, unwind_fields=q.unwind)
    return dict(result=rv[q.start_idx:])